;;; Procedures modeled on those from various CHICKEN libraries,
;;; (chicken string) in particular.
;;;
;;; These are extensions to SRFI 207.

(define (%to-byte x)
  (cond ((exact-natural? x) x)
        ((and (char? x) (char<=? x #\delete)) (char->integer x))
        (else (error "invalid bytestring element" x))))

(: bytestring-translate
   (bytestring (or char fixnum list) (or char fixnum list) -> bytestring))
(define (bytestring-translate bstring from to)
  (assert (or (char? from) (exact-natural? from) (null? from) (pair? from)))
  (assert (or (char? to) (exact-natural? to) (null? from) (pair? to)))
  (assert (if (pair? from)
              (and (pair? to) (<= (length from) (length to)))
              #t)
          "`to` must contain at least as many elements as `from`"
          from
          to)
  (let ((index (%make-indexer from))
        (tos (if (or (null? to) (pair? to)) to (list to))))
    (u8vector-map (lambda (b)
                    (cond ((index b) =>
                           (lambda (j) (%to-byte (list-ref tos j))))
                          (else b)))
                  bstring)))

(define (%make-indexer obj)
  (cond ((integer? obj)
         (assert (exact-natural? obj) "invalid byte" obj)
         (lambda (b) (and (= obj b) 0)))
        ((char? obj)
         (assert (char<=? obj #\delete) "invalid character" obj)
         (lambda (b) (and (= b (char->integer obj)) 0)))
        ((null? obj) (lambda (b) #f))
        ((pair? obj)
         (lambda (b)
           (list-index (lambda (c) (= b (%to-byte c))) obj)))
        (else (error "invalid 'from' object" obj))))

;; Based on string-translate*, but simpler: only maps bytes to bytes.
;; tmap may mix bytes and characters.
(: bytestring-substitute (bytestring (list-of pair) -> bytestring))
(define (bytestring-substitute bstring tmap)
  (assert (bytevector? bstring))
  (assert (or (pair? tmap) (null? tmap)))
  (let ((tmap-bytes (map (lambda (p)
                           (cons (%to-byte (car p))
                                 (%to-byte (cdr p))))
                         tmap)))
    (u8vector-map (lambda (b)
                    (cond ((assv b tmap-bytes) => cdr)
                          (else b)))
                  bstring)))

(: bytestring-chomp
   (bytestring #!optional bytestring -> bytestring))
(define bytestring-chomp
  (case-lambda
    ((bstring) (bytestring-chomp bstring #u8(#xa)))
    ((bstring suffix)
     (assert (bytevector? bstring))
     (assert (bytevector? suffix))
     (let ((blen (bytevector-length bstring))
           (slen (bytevector-length suffix)))
       (if (bytestring-suffix? suffix bstring)
           (bytevector-copy bstring 0 (- blen slen))
           (bytevector-copy bstring))))))

(: subbytestring=?
   (bytestring bytestring #!optional integer integer integer --> boolean))
(define (subbytestring=? bs1 bs2 . opt)
  (assert (bytevector? bs1))
  (assert (bytevector? bs2))
  (if (null? opt)
      (or (eq? bs1 bs2)    ; quick check
          (equal? bs1 bs2))
      (let-optionals* opt ((start1 0)
                           (start2 0)
                           (len (- (bytevector-length bs1) start1)))
        (%subbytestring=? = bs1 bs2 start1 start2 len))))

(: subbytestring-ci=?
   (bytestring bytestring #!optional integer integer integer --> boolean))
(define (subbytestring-ci=? bs1 bs2 . opt)
  (assert (bytevector? bs1))
  (assert (bytevector? bs2))
  (or (and (null? opt) (eq? bs1 bs2))    ; quick check
      (let-optionals* opt ((start1 0)
                           (start2 0)
                           (len (- (bytevector-length bs1) start1)))
        (%subbytestring=? %byte-ci=? bs1 bs2 start1 start2 len))))

;; Could be optimized.
(define (%byte-ci=? b c)
  (or (= b c)
      (and (<= b #x7f)
           (<= c #x7f)
           (char-ci=? (integer->char b) (integer->char c)))))

(define (%subbytestring=? comp bs1 bs2 start1 start2 len)
  (let ((len1 (bytevector-length bs1))
        (len2 (bytevector-length bs2)))
    (assert (and (exact-natural? start1) (< start1 len1))
            "invalid start index" start1 bs1)
    (assert (and (exact-natural? start2) (< start2 len2))
            "invalid start index" start2 bs2)
    (assert (exact-natural? len)
            "invalid substring length" len bs1 start1 bs2 start2)
    (and (<= len len1)
         (<= len len2)
         (let loop ((i start1) (j start2) (k 0))
           (cond ((= k len) #t)
                 ((or (>= i len1) (>= j len2)) #f)
                 ((comp (bytevector-u8-ref bs1 i)
                        (bytevector-u8-ref bs2 j))
                  (loop (+ i 1) (+ j 1) (+ k 1)))
                 (else #f))))))

(: bytestring-compare3 (bytevector bytevector --> integer))
(define (bytestring-compare3 bs1 bs2)
  (assert (bytevector? bs1))
  (assert (bytevector? bs2))
  (%bytestring-compare bs1 bs2 -1 0 1))

(: bytestring-compare3-ci (bytevector bytevector --> integer))
(define (bytestring-compare3-ci bs1 bs2)
  (assert (bytevector? bs1))
  (assert (bytevector? bs2))
  (%bytestring-compare-ci bs1 bs2 -1 0 1))
