;;; Bytestring procedures derived from SRFI 152.  Many implementations
;;; are based on SRFI 178 adaptations of the SRFI 1 sample impl..

;;; TODO: start/end indices.

(: bytestring-prefix-length (bytevector bytevector --> integer))
(define (bytestring-prefix-length bvec1 bvec2)
  (assert (bytevector? bvec1))
  (assert (bytevector? bvec2))
  (let ((end (min (bytevector-length bvec1) (bytevector-length bvec2))))
    (if (eq? bvec1 bvec2)
        end
        (let lp ((i 0))
          (if (or (>= i end)
                  (not (= (bytevector-u8-ref bvec1 i)
                          (bytevector-u8-ref bvec2 i))))
              i
              (lp (+ i 1)))))))

(: bytestring-suffix-length (bytevector bytevector --> integer))
(define (bytestring-suffix-length bvec1 bvec2)
  (assert (bytevector? bvec1))
  (assert (bytevector? bvec2))
  (let ((end1 (bytevector-length bvec1))
        (end2 (bytevector-length bvec2)))
    (let* ((delta (min end1 end2))
           (start (- end1 delta)))
      (if (eq? bvec1 bvec2)
          delta
          (let lp ((i (- end1 1)) (j (- end2 1)))
            (if (or (< i start)
                    (not (= (bytevector-u8-ref bvec1 i)
                            (bytevector-u8-ref bvec2 j))))
                (- (- end1 i) 1)
                (lp (- i 1) (- j 1))))))))

(: bytestring-prefix? (bytevector bytevector --> boolean))
(define (bytestring-prefix? bvec1 bvec2)
  (assert (bytevector? bvec1))
  (assert (bytevector? bvec2))
  (let ((len1 (bytevector-length bvec1)))
    (and (<= len1 (bytevector-length bvec2))
         (= (bytestring-prefix-length bvec1 bvec2) len1))))

(: bytestring-suffix? (bytevector bytevector --> boolean))
(define (bytestring-suffix? bvec1 bvec2)
  (assert (bytevector? bvec1))
  (assert (bytevector? bvec2))
  (let ((len1 (bytevector-length bvec1)))
    (and (<= len1 (bytevector-length bvec2))
         (= (bytestring-suffix-length bvec1 bvec2) len1))))

;; Known as "chop" in (chicken string).
(: bytestring-segment (bytevector integer -> (list-of bytevector)))
(define (bytestring-segment bvec k)
  (assert (bytevector? bvec))
  (assert (and (exact-integer? k) (>= k 1))
          "segment length must be a positive integer" k)
  (let ((len (bytevector-length bvec)))
    (unfold (lambda (i) (>= i len))
            (lambda (i) (bytevector-copy bvec i (min (+ i k) len)))
            (lambda (i) (+ i k))
            0)))

;; FIXME: Inefficient.  We should correctly implement
;; Knuth-Morris-Pratt searching and rewrite this.
(: bytestring-contains (bytevector bytevector #!optional integer
   integer integer integer --> (or integer false)))
(define (bytestring-contains bs1 bs2 . opt)
  (assert (bytevector? bs1))
  (assert (bytevector? bs2))
  (let ((len1 (bytevector-length bs1))
        (len2 (bytevector-length bs2)))
    (let-optionals opt ((start1 0)
                        (end1 len1)
                        (start2 0)
                        (end2 len2))
      (assert (exact-natural? start1))
      (assert (exact-natural? end1))
      (assert (exact-natural? start2))
      (assert (exact-natural? end2))
      (assert (<= 0 start1 end1 len1)
              "invalid start/end indices" bs1 start1 end1)
      (assert (<= 0 start2 end2 len2)
              "invalid start/end indices" bs2 start2 end2)
      (if (= start2 end2)
          start1
          (let* ((sub-len (- end2 start2))
                 (i-bound (- end1 sub-len)))
            (let loop ((i start1))
              (and (<= i i-bound)
                   (if (subbytestring=? bs1 bs2 i start2 sub-len)
                       i
                       (loop (+ i 1))))))))))

(: bytestring-contains-right (bytevector bytevector #!optional integer
   integer integer integer --> (or integer false)))
(define (bytestring-contains-right bs1 bs2 . opt)
  (assert (bytevector? bs1))
  (assert (bytevector? bs2))
  (let ((len1 (bytevector-length bs1))
        (len2 (bytevector-length bs2)))
    (let-optionals opt ((start1 0)
                        (end1 len1)
                        (start2 0)
                        (end2 len2))
      (assert (exact-natural? start1))
      (assert (exact-natural? end1))
      (assert (exact-natural? start2))
      (assert (exact-natural? end2))
      (assert (<= 0 start1 end1 len1)
              "invalid start/end indices" bs1 start1 end1)
      (assert (<= 0 start2 end2 len2)
              "invalid start/end indices" bs2 start2 end2)
      (if (= start2 end2)
          end1
          (let* ((sub-len (- end2 start2))
                 (i-bound start1))
            (let loop ((i (- end1 sub-len)))
              (and (>= i i-bound)
                   (if (subbytestring=? bs1 bs2 i start2 sub-len)
                       i
                       (loop (- i 1))))))))))

(: bytestring-concatenate-reverse
   ((list-of bytevector) #!optional bytevector integer -> bytevector))
(define (bytestring-concatenate-reverse bss . opt)
  (let-optionals* opt ((final #u8())
                       (end (bytevector-length final)))
    (assert (or (pair? bss) (null? bss)))
    (assert (bytevector? final))
    (assert (exact-integer? end))
    (assert (<= 0 end (bytevector-length final))
            "end index out of range" final end)
    (let ((len (fold + 0 (map bytevector-length bss))))
      (%concat-reverse len bss final end))))

(: %concat-reverse
   (integer (list-of bytevector) bytevector integer -> bytevector))
(define (%concat-reverse len bss final end)
  (let ((res (make-bytevector (+ len end))))
    (unless (zero? end)
      (bytevector-copy! res len final 0 end))
    (fold (lambda (bs i)
            (assert (bytevector? bs))
            (let* ((l (bytevector-length bs))
                   (i* (- i l)))
              (bytevector-copy! res i* bs 0 l)
              i*))
          len
          bss)
    res))

(: bytestring-replicate
   (bytevector integer integer #!optional integer integer -> bytevector))
(define (bytestring-replicate bvec from to . opt)
  (assert (bytevector? bvec))
  (assert (exact-integer? from))
  (assert (exact-integer? to))
  (assert (<= from to) "invalid from/to indices" from to)
  (let ((full-len (bytevector-length bvec)))
    (let-optionals opt ((start 0) (end full-len))
      (assert (exact-integer? start))
      (assert (exact-integer? end))
      (assert (<= 0 start end full-len)
              "start/end indices out of range" bvec start end)
      (let ((len (- end start)) (reslen (- to from)))
        (cond ((zero? reslen) #u8())  ; OK if end = start here.
              ((= len 1)              ; Fast path.
               (make-bytevector reslen (bytevector-u8-ref bvec start)))
              ((= (floor (/ from len)) (floor (/ to len)))
               (bytevector-copy bvec  ; All within one span.
                                (+ start (modulo from len))
                                (+ start (modulo to len))))
              (else
               (assert (positive? len)
                       "invalid substring length" start end)
               (let ((res (make-bytevector reslen)))
                 (%multispan-repcopy! res 0 bvec from to start end)
                 res)))))))

;; Adapted from SRFI 13.  Thanks, Olin!
(: %multispan-repcopy!
   (bytevector integer bytevector integer integer integer integer ->
    bytevector))
(define (%multispan-repcopy! target tstart bs from to start end)
  (let* ((len (- end start))
         (i0 (+ start (modulo from len)))
         (total-chars (- to from)))
    ;; Copy the initial, partial span.
    (bytevector-copy! target tstart bs i0 end)
    (let* ((ncopied (- end i0))             ; # of chars copied so far.
           (nleft (- total-chars ncopied))  ; # of chars left to copy.
           (nspans (quotient nleft len)))   ; # of whole spans to copy.
      ;; Copy the central, whole spans.
      (do ((i (+ tstart ncopied) (+ i len))   ; Current target index.
           (nspans nspans (- nspans 1)))      ; # of spans to copy.
          ((zero? nspans)
           ;; Copy the ending partial span.
           (let ((frag-end (+ start (- total-chars (- i tstart)))))
             (bytevector-copy! target i bs start frag-end)))
        ;; Copy a whole span.
        (bytevector-copy! target i bs start end)))))

;;;; Case-insensitive comparisons

;;; These could be tuned.
(: %byte-ci=? (integer integer --> boolean))
(define (%byte-ci=? b c)
  (char-ci=? (integer->char b) (integer->char c)))

(: %byte-ci<? (integer integer --> boolean))
(define (%byte-ci<? b c)
  (char-ci<? (integer->char b) (integer->char c)))

(: %bytestring-prefix-length-ci (bytevector bytevector --> integer))
(define (%bytestring-prefix-length-ci bvec1 bvec2)
  (let ((end (min (bytevector-length bvec1)
                  (bytevector-length bvec2))))
    (if (eq? bvec1 bvec2)  ; fast path
        end
        (let lp ((i 0))
          (if (or (>= i end)
                  (not (%byte-ci=? (bytevector-u8-ref bvec1 i)
                                   (bytevector-u8-ref bvec2 i))))
              i
              (lp (+ i 1)))))))

(: %bytestring-compare-ci (bytevector bytevector * * * --> *))
(define (%bytestring-compare-ci bvec1 bvec2 res< res= res>)
  (let ((len1 (bytevector-length bvec1))
        (len2 (bytevector-length bvec2)))
    (let ((match (%bytestring-prefix-length-ci bvec1 bvec2)))
      (if (= match len1)
          (if (= match len2) res= res<)
          (if (= match len2)
              res>
              (if (%byte-ci<? (bytevector-u8-ref bvec1 match)
                              (bytevector-u8-ref bvec2 match))
                  res<
                  res>))))))

(: bytestring-ci=? (bytevector bytevector #!rest bytevector --> boolean))
(define (bytestring-ci=? bvec1 bvec2 . rest)
  (assume (bytevector? bvec1))
  (assume (bytevector? bvec2))
  (if (null? rest)                ; fast path
      (or (eq? bvec1 bvec2)       ; quick check
          (%bytestring-compare-ci bvec1 bvec2 #f #t #f))
      (let loop ((bs bvec1) (bss (cons bvec2 rest)))  ; variadic path
        (or (null? bss)
            (let ((bs* (car bss)))
              (and (or (eq? bs bs*)
                       (%bytestring-compare-ci bs bs* #f #t #f))
                   (loop bs* (cdr bss))))))))

(: bytestring-ci<? (bytevector bytevector #!rest bytevector --> boolean))
(define (bytestring-ci<? bvec1 bvec2 . rest)
  (assume (bytevector? bvec1))
  (assume (bytevector? bvec2))
  (if (null? rest)                       ; fast path
      (and (not (eq? bvec1 bvec2))       ; quick check
           (%bytestring-compare-ci bvec1 bvec2 #t #f #f))
      (let loop ((bs bvec1) (bss (cons bvec2 rest)))  ; variadic path
        (or (null? bss)
            (let ((bs* (car bss)))
              (and (not (eq? bs bs*))
                   (%bytestring-compare-ci bs bs* #t #f #f)
                   (loop bs* (cdr bss))))))))

(: bytestring-ci>? (bytevector bytevector #!rest bytevector --> boolean))
(define (bytestring-ci>? bvec1 bvec2 . rest)
  (assume (bytevector? bvec1))
  (assume (bytevector? bvec2))
  (if (null? rest)                       ; fast path
      (and (not (eq? bvec1 bvec2))       ; quick check
           (%bytestring-compare-ci bvec1 bvec2 #f #f #t))
      (let loop ((bs bvec1) (bss (cons bvec2 rest)))  ; variadic path
        (or (null? bss)
            (let ((bs* (car bss)))
              (and (not (eq? bs bs*))
                   (%bytestring-compare-ci bs bs* #f #f #t)
                   (loop bs* (cdr bss))))))))

(: bytestring-ci<=? (bytevector bytevector #!rest bytevector --> boolean))
(define (bytestring-ci<=? bvec1 bvec2 . rest)
  (assume (bytevector? bvec1))
  (assume (bytevector? bvec2))
  (if (null? rest)                       ; fast path
      (or (eq? bvec1 bvec2)              ; quick check
          (%bytestring-compare-ci bvec1 bvec2 #t #t #f))
      (let loop ((bs bvec1) (bss (cons bvec2 rest)))  ; variadic path
        (or (null? bss)
            (let ((bs* (car bss)))
              (and (or (eq? bs bs*)
                       (%bytestring-compare-ci bs bs* #t #t #f))
                   (loop bs* (cdr bss))))))))

(: bytestring-ci>=? (bytevector bytevector #!rest bytevector --> boolean))
(define (bytestring-ci>=? bvec1 bvec2 . rest)
  (assume (bytevector? bvec1))
  (assume (bytevector? bvec2))
  (if (null? rest)                       ; fast path
      (or (eq? bvec1 bvec2)              ; quick check
          (%bytestring-compare-ci bvec1 bvec2 #f #t #t))
      (let loop ((bs bvec1) (bss (cons bvec2 rest)))  ; variadic path
        (or (null? bss)
            (let ((bs* (car bss)))
              (and (or (eq? bs bs*)
                       (%bytestring-compare-ci bs bs* #f #t #t))
                   (loop bs* (cdr bss))))))))

;;; TODO: start/end indices.

(: bytestring-prefix-length-ci (bytevector bytevector --> integer))
(define (bytestring-prefix-length-ci bvec1 bvec2)
  (assert (bytevector? bvec1))
  (assert (bytevector? bvec2))
  (%bytestring-prefix-length-ci bvec1 bvec2))

(: bytestring-prefix-ci? (bytevector bytevector --> boolean))
(define (bytestring-prefix-ci? bvec1 bvec2)
  (assert (bytevector? bvec1))
  (assert (bytevector? bvec2))
  (let ((len1 (bytevector-length bvec1)))
    (and (<= len1 (bytevector-length bvec2))
         (= (%bytestring-prefix-length-ci bvec1 bvec2) len1))))

(: %bytestring-suffix-length-ci (bytevector bytevector --> integer))
(define (%bytestring-suffix-length-ci bvec1 bvec2)
  (let ((end1 (bytevector-length bvec1))
        (end2 (bytevector-length bvec2)))
    (let* ((delta (min end1 end2))
           (start (- end1 delta)))
      (if (eq? bvec1 bvec2)
          delta
          (let lp ((i (- end1 1)) (j (- end2 1)))
            (if (or (< i start)
                    (not (%byte-ci=? (bytevector-u8-ref bvec1 i)
                                     (bytevector-u8-ref bvec2 j))))
                (- (- end1 i) 1)
                (lp (- i 1) (- j 1))))))))

(: bytestring-suffix-length-ci (bytevector bytevector --> integer))
(define (bytestring-suffix-length-ci bvec1 bvec2)
  (assert (bytevector? bvec1))
  (assert (bytevector? bvec2))
  (%bytestring-suffix-length-ci bvec1 bvec2))

(: bytestring-suffix-ci? (bytevector bytevector --> boolean))
(define (bytestring-suffix-ci? bvec1 bvec2)
  (assert (bytevector? bvec1))
  (assert (bytevector? bvec2))
  (let ((len1 (bytevector-length bvec1)))
    (and (<= len1 (bytevector-length bvec2))
         (= len1 (%bytestring-suffix-length-ci bvec1 bvec2)))))
